import io
import json
import os
from pathlib import Path
from uuid import UUID
import zipfile
from typing import Optional

from dotenv import load_dotenv
from pydash import snake_case

from vellum.client.core.api_error import ApiError
from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.utils.uuid import is_valid_uuid
from vellum.workflows.vellum_client import create_vellum_client
from vellum_cli.config import (
    DEFAULT_WORKSPACE_CONFIG,
    VellumCliConfig,
    WorkflowConfig,
    WorkflowDeploymentConfig,
    load_vellum_cli_config,
)
from vellum_cli.logger import handle_cli_error, load_cli_logger

ERROR_LOG_FILE_NAME = "error.log"
METADATA_FILE_NAME = "metadata.json"


class WorkflowConfigResolutionResult(UniversalBaseModel):
    workflow_config: Optional[WorkflowConfig] = None
    pk: Optional[str] = None


class RunnerConfig(UniversalBaseModel):
    container_image_name: Optional[str] = None
    container_image_tag: Optional[str] = None


class PullContentsMetadata(UniversalBaseModel):
    label: Optional[str] = None
    runner_config: Optional[RunnerConfig] = None
    deployment_id: Optional[UUID] = None
    deployment_name: Optional[str] = None
    workflow_sandbox_id: Optional[UUID] = None


def _resolve_workflow_config(
    config: VellumCliConfig,
    module: Optional[str] = None,
    workflow_sandbox_id: Optional[str] = None,
    workflow_deployment: Optional[str] = None,
    workspace: Optional[str] = None,
) -> WorkflowConfigResolutionResult:
    if workflow_sandbox_id and workflow_deployment:
        raise ValueError("Cannot specify both workflow_sandbox_id and workflow_deployment")

    if module:
        workflow_config = next((w for w in config.workflows if w.module == module), None)
        if not workflow_config and workflow_sandbox_id:
            workflow_config = WorkflowConfig(
                workflow_sandbox_id=workflow_sandbox_id,
                module=module,
                workspace=workspace or DEFAULT_WORKSPACE_CONFIG.name,
            )
            config.workflows.append(workflow_config)
            return WorkflowConfigResolutionResult(
                workflow_config=workflow_config,
                pk=workflow_sandbox_id,
            )

        return WorkflowConfigResolutionResult(
            workflow_config=workflow_config,
            pk=workflow_config.workflow_sandbox_id if workflow_config else None,
        )
    elif workflow_sandbox_id:
        workflow_config = next((w for w in config.workflows if w.workflow_sandbox_id == workflow_sandbox_id), None)
        if workflow_config:
            return WorkflowConfigResolutionResult(
                workflow_config=workflow_config,
                pk=workflow_sandbox_id,
            )

        # We use an empty module name to indicate that we want to backfill it once we have the Workflow Sandbox Label
        workflow_config = WorkflowConfig(
            workflow_sandbox_id=workflow_sandbox_id,
            module="",
        )
        config.workflows.append(workflow_config)
        return WorkflowConfigResolutionResult(
            workflow_config=workflow_config,
            pk=workflow_config.workflow_sandbox_id,
        )
    elif workflow_deployment:
        if is_valid_uuid(workflow_deployment):
            # name may also be a valid UUID
            workflow_config = next(
                (
                    w
                    for w in config.workflows
                    if w.deployments
                    and (
                        str(w.deployments[0].id) == workflow_deployment or w.deployments[0].name == workflow_deployment
                    )
                ),
                None,
            )
        else:
            workflow_config = next(
                (w for w in config.workflows if w.deployments and w.deployments[0].name == workflow_deployment), None
            )
        if workflow_config:
            return WorkflowConfigResolutionResult(
                workflow_config=workflow_config,
                pk=workflow_deployment,
            )

        workflow_config = WorkflowConfig(
            module="",
        )
        config.workflows.append(workflow_config)
        return WorkflowConfigResolutionResult(
            workflow_config=workflow_config,
            pk=workflow_deployment,
        )
    elif config.workflows:
        return WorkflowConfigResolutionResult(
            workflow_config=config.workflows[0],
            pk=config.workflows[0].workflow_sandbox_id,
        )

    return WorkflowConfigResolutionResult()


def pull_command(
    module: Optional[str] = None,
    workflow_sandbox_id: Optional[str] = None,
    workflow_deployment: Optional[str] = None,
    include_json: Optional[bool] = None,
    exclude_code: Optional[bool] = None,
    strict: Optional[bool] = None,
    include_sandbox: Optional[bool] = None,
    target_directory: Optional[str] = None,
    workspace: Optional[str] = None,
) -> None:
    load_dotenv(dotenv_path=os.path.join(os.getcwd(), ".env"))
    logger = load_cli_logger()
    config = load_vellum_cli_config()

    workflow_config_result = _resolve_workflow_config(
        config=config,
        module=module,
        workflow_sandbox_id=workflow_sandbox_id,
        workflow_deployment=workflow_deployment,
        workspace=workspace,
    )

    workflow_config = workflow_config_result.workflow_config
    if not workflow_config:
        raise ValueError("No workflow config found in project to pull from.")

    pk = workflow_config_result.pk
    if not pk:
        raise ValueError("No workflow sandbox ID found in project to pull from.")

    if workflow_config.module:
        logger.info(f"Pulling workflow {workflow_config.module}...")
    else:
        logger.info(f"Pulling workflow from {pk}...")

    resolved_workspace = workspace or workflow_config.workspace or DEFAULT_WORKSPACE_CONFIG.name
    workspace_config = (
        next((w for w in config.workspaces if w.name == resolved_workspace), DEFAULT_WORKSPACE_CONFIG)
        if workspace
        else DEFAULT_WORKSPACE_CONFIG
    )
    api_key = os.getenv(workspace_config.api_key)
    if not api_key:
        raise ValueError(f"No API key value found in environment for workspace '{workspace_config.name}'.")

    client = create_vellum_client(
        api_key=api_key,
        api_url=workspace_config.api_url,
    )
    query_parameters = {}

    if include_json:
        query_parameters["include_json"] = include_json
    if exclude_code:
        query_parameters["exclude_code"] = exclude_code
    if strict:
        query_parameters["strict"] = strict
    if include_sandbox:
        query_parameters["include_sandbox"] = include_sandbox

    response = client.workflows.pull(
        pk,
        request_options={"additional_query_parameters": query_parameters},
    )

    try:
        zip_bytes = b"".join(response)
    except ApiError as e:
        if e.status_code == 401 or e.status_code == 403:
            handle_cli_error(
                logger,
                title="Authentication failed",
                message="Unable to authenticate with the Vellum API.",
                suggestion="Please make sure your `VELLUM_API_KEY` environment variable is set correctly and that you have access to this workflow.",  # noqa: E501
            )

        if e.status_code == 404:
            handle_cli_error(
                logger,
                title="Workflow not found",
                message=f"The workflow with ID '{pk}' could not be found.",
                suggestion="Please verify the workflow ID is correct and that you have access to it in your workspace.",
            )

        if e.status_code == 500:
            handle_cli_error(
                logger,
                title="Server error occurred",
                message="The Vellum API encountered an internal server error while processing your request.",
                suggestion="Please try again in a few moments. If the problem persists, contact Vellum support with the workflow ID and timestamp.",  # noqa: E501
            )

        if e.status_code == 502 or e.status_code == 503 or e.status_code == 504:
            handle_cli_error(
                logger,
                title="Service temporarily unavailable",
                message="The Vellum API is temporarily unavailable or experiencing high load.",
                suggestion="Please wait a moment and try again. If the issue continues, check the Vellum status page or contact support.",  # noqa: E501
            )

        handle_cli_error(
            logger,
            title="API request failed",
            message=f"The API request failed with status code {e.status_code}.",
            suggestion="Please verify your `VELLUM_API_URL` environment variable is set correctly and try again.",
        )

    zip_buffer = io.BytesIO(zip_bytes)
    error_content = ""
    try:
        with zipfile.ZipFile(zip_buffer) as zip_file:
            if METADATA_FILE_NAME in zip_file.namelist():
                metadata_json: Optional[dict] = None
                with zip_file.open(METADATA_FILE_NAME) as source:
                    metadata_json = json.load(source)

                pull_contents_metadata = PullContentsMetadata.model_validate(metadata_json)

                if pull_contents_metadata.runner_config:
                    workflow_config.container_image_name = pull_contents_metadata.runner_config.container_image_name
                    workflow_config.container_image_tag = pull_contents_metadata.runner_config.container_image_tag
                    if workflow_config.container_image_name and not workflow_config.container_image_tag:
                        workflow_config.container_image_tag = "latest"
                if not workflow_config.workflow_sandbox_id and pull_contents_metadata.workflow_sandbox_id:
                    workflow_config.workflow_sandbox_id = str(pull_contents_metadata.workflow_sandbox_id)
                if not workflow_config.module and workflow_deployment and pull_contents_metadata.deployment_name:
                    workflow_config.module = snake_case(pull_contents_metadata.deployment_name)
                if not workflow_config.module and pull_contents_metadata.label:
                    workflow_config.module = snake_case(pull_contents_metadata.label)

                # Save or update the deployment info when pulling with --workflow-deployment
                if workflow_deployment:
                    workflow_deployment_id = pull_contents_metadata.deployment_id
                    existing_deployment = next(
                        (d for d in workflow_config.deployments if d.id == workflow_deployment_id), None
                    )

                    if existing_deployment:
                        if pull_contents_metadata.label:
                            existing_deployment.label = pull_contents_metadata.label
                    else:
                        deployment_config = WorkflowDeploymentConfig(
                            id=workflow_deployment_id,
                            label=pull_contents_metadata.label,
                            name=pull_contents_metadata.deployment_name,
                        )
                        workflow_config.deployments.append(deployment_config)

            if not workflow_config.module:
                raise ValueError(f"Failed to resolve a module name for Workflow {pk}")

            # Use target_directory if provided, otherwise use current working directory
            base_dir = os.path.join(os.getcwd(), target_directory) if target_directory else os.getcwd()
            target_dir = os.path.join(base_dir, *workflow_config.module.split("."))
            workflow_config.target_directory = target_dir if target_directory else None

            # Delete files in target_dir that aren't in the zip file
            if os.path.exists(target_dir):
                ignore_patterns = (
                    workflow_config.ignore
                    if isinstance(workflow_config.ignore, list)
                    else [workflow_config.ignore] if isinstance(workflow_config.ignore, str) else []
                )
                existing_files = []
                for root, _, files in os.walk(target_dir):
                    for file in files:
                        rel_path = os.path.relpath(os.path.join(root, file), target_dir)
                        existing_files.append(rel_path)

                for file in existing_files:
                    if any(Path(file).match(ignore_pattern) for ignore_pattern in ignore_patterns):
                        continue

                    if file not in zip_file.namelist():
                        file_path = os.path.join(target_dir, file)
                        logger.info(f"Deleting {file_path}...")
                        os.remove(file_path)

            for file_name in zip_file.namelist():
                with zip_file.open(file_name) as source:
                    content = source.read().decode("utf-8")
                    if file_name == ERROR_LOG_FILE_NAME:
                        error_content = content
                        continue
                    if file_name == METADATA_FILE_NAME:
                        continue

                    target_file = os.path.join(target_dir, file_name)
                    os.makedirs(os.path.dirname(target_file), exist_ok=True)
                    with open(target_file, "w") as target:
                        logger.info(f"Writing to {target_file}...")
                        target.write(content)
    except zipfile.BadZipFile:
        handle_cli_error(
            logger,
            title="Invalid response format",
            message="The API returned an invalid zip file format.",
            suggestion="Please verify your `VELLUM_API_URL` environment variable is set correctly and try again.",
        )

    if include_json:
        logger.warning(
            """The pulled JSON representation of the Workflow should be used for debugging purposely only. \
Its schema should be considered unstable and subject to change at any time."""
        )

    if include_sandbox:
        if not workflow_config.ignore:
            workflow_config.ignore = "sandbox.py"
        elif isinstance(workflow_config.ignore, str) and "sandbox.py" != workflow_config.ignore:
            workflow_config.ignore = [workflow_config.ignore, "sandbox.py"]
        elif isinstance(workflow_config.ignore, list) and "sandbox.py" not in workflow_config.ignore:
            workflow_config.ignore.append("sandbox.py")

    config.save()

    if error_content:
        logger.error(error_content)
    else:
        logger.info(f"Successfully pulled Workflow into {target_dir}")



import datetime
import itertools
import json
import random
import re
from typing import Any, Callable, Dict, Union

import dateutil
import pydash
import pytz
import yaml

from vellum.utils.templating.custom_filters import is_valid_json_string, replace, safe_tojson

DEFAULT_JINJA_GLOBALS: Dict[str, Any] = {
    "datetime": datetime,
    "dateutil": dateutil,
    "itertools": itertools,
    "json": json,
    "pydash": pydash,
    "pytz": pytz,
    "random": random,
    "re": re,
    "yaml": yaml,
}

FilterFunc = Union[
    Callable[[Union[str, bytes]], bool],  # is_valid_json_string
    Callable[[Any, Any, Any], str],  # replace
    Callable[[Any], str],  # safe_tojson
]

DEFAULT_JINJA_CUSTOM_FILTERS: Dict[str, FilterFunc] = {
    "is_valid_json_string": is_valid_json_string,
    "replace": replace,
    "tojson": safe_tojson,
}



from typing import Any, Dict, Generic, Iterator, Type, Union

from vellum.workflows.constants import undefined
from vellum.workflows.errors import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.nodes.displayable.bases import BaseInlinePromptNode as BaseInlinePromptNode
from vellum.workflows.nodes.displayable.bases.utils import process_additional_prompt_outputs
from vellum.workflows.outputs import BaseOutput
from vellum.workflows.types import MergeBehavior
from vellum.workflows.types.generics import StateType


class InlinePromptNode(BaseInlinePromptNode[StateType], Generic[StateType]):
    """
    Used to execute a Prompt defined inline.

    prompt_inputs: EntityInputsInterface - The inputs for the Prompt
    ml_model: str - Either the ML Model's UUID or its name.
    blocks: List[PromptBlock] - The blocks that make up the Prompt
    functions: Optional[List[FunctionDefinition]] - The functions to include in the Prompt
    parameters: PromptParameters - The parameters for the Prompt
    expand_meta: Optional[AdHocExpandMeta] - Expandable execution fields to include in the response
    request_options: Optional[RequestOptions] - The request options to use for the Prompt Execution
    """

    class Trigger(BaseInlinePromptNode.Trigger):
        merge_behavior = MergeBehavior.AWAIT_ANY

    class Outputs(BaseInlinePromptNode.Outputs):
        """
        The outputs of the InlinePromptNode.

        text: str - The result of the Prompt Execution
        json: Optional[Dict[Any, Any]] - The result of the Prompt Execution in JSON format
        """

        text: str
        json: Union[Dict[Any, Any], Type[undefined]] = undefined

    def run(self) -> Iterator[BaseOutput]:
        outputs = yield from self._process_prompt_event_stream()
        if not outputs:
            raise NodeException(
                message="Expected to receive outputs from Prompt",
                code=WorkflowErrorCode.INTERNAL_ERROR,
            )

        text_output, json_output = process_additional_prompt_outputs(outputs)
        yield BaseOutput(name="text", value=text_output)

        if json_output:
            yield BaseOutput(name="json", value=json_output)



from vellum import PromptParameters

DEFAULT_PROMPT_PARAMETERS = PromptParameters(
    stop=[],
    temperature=0.0,
    max_tokens=4096,
    top_p=1.0,
    top_k=0,
    frequency_penalty=0.0,
    presence_penalty=0.0,
    logit_bias=None,
    custom_parameters=None,
)



import pytest
from typing import Any, List, Union

from pydantic import BaseModel

from vellum.client.types.chat_message import ChatMessage
from vellum.client.types.function_call import FunctionCall
from vellum.client.types.vellum_value import VellumValue
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.nodes.utils import cast_to_output_type, parse_type_from_str
from vellum.workflows.types.core import Json


class Person(BaseModel):
    name: str
    age: int
    colors: List[str]


@pytest.mark.parametrize(
    "input_str, output_type, expected_result",
    [
        ("hello", str, "hello"),
        ("3.14", float, 3.14),
        ("42", int, 42),
        ("True", bool, True),
        ("", List[str], []),  # Empty string should return an empty list
        ("[1, 2, 3]", List[int], [1, 2, 3]),
        ('["a", "b", "c"]', List[str], ["a", "b", "c"]),
        ('{"name": "Alice", "age": 30}', Json, {"name": "Alice", "age": 30}),
        (
            '{"type": "FUNCTION_CALL", "value": {"name": "test", "args": [1, 2]}}',
            Json,
            {"name": "test", "args": [1, 2]},
        ),
        ("42", Union[int, str], 42),
        ("hello", Union[int, str], "hello"),
    ],
    ids=[
        "str",
        "float",
        "int",
        "bool",
        "empty_list",
        "list_of_int",
        "list_of_str",
        "simple_json",
        "function_call_json",
        "union_int",
        "union_str",
    ],
)
def test_parse_type_from_str_basic_cases(input_str, output_type, expected_result):
    result = parse_type_from_str(input_str, output_type)
    assert result == expected_result


def test_parse_type_from_str_pydantic_models():
    person_json = '{"name": "Alice", "age": 30, "colors": ["red", "blue"]}'
    person = parse_type_from_str(person_json, Person)
    assert isinstance(person, Person)
    assert person.name == "Alice"
    assert person.age == 30
    assert person.colors == ["red", "blue"]


def test_parse_type_from_str_list_of_models():
    person_list_json = '[{"name": "Alice", "age": 30, "colors": ["red", "blue"]}, {"name": "Bob", "age": 25, "colors": ["green", "yellow"]}]'  # noqa: E501
    persons = parse_type_from_str(person_list_json, List[Person])
    assert len(persons) == 2
    assert all(isinstance(p, Person) for p in persons)
    assert persons[0].name == "Alice"
    assert persons[0].age == 30
    assert persons[0].colors == ["red", "blue"]
    assert persons[1].name == "Bob"
    assert persons[1].age == 25
    assert persons[1].colors == ["green", "yellow"]


@pytest.mark.parametrize(
    "input_str, output_type, expected_exception, expected_code, expected_message",
    [
        (
            "{invalid json}",
            List[str],
            NodeException,
            WorkflowErrorCode.INVALID_OUTPUTS,
            "Invalid JSON Array format for result_as_str",
        ),
        (
            "{invalid json}",
            Person,
            NodeException,
            WorkflowErrorCode.INVALID_OUTPUTS,
            "Invalid JSON format for result_as_str",
        ),
        (
            "{invalid json}",
            Json,
            NodeException,
            WorkflowErrorCode.INVALID_OUTPUTS,
            "Invalid JSON format for result_as_str",
        ),
        ('{"name": "Alice"}', List[str], ValueError, None, "Expected a list of items for result_as_str, received dict"),
        ("data", object, ValueError, None, "Unsupported output type: <class 'object'>"),
    ],
    ids=[
        "invalid_json_list",
        "invalid_json_model",
        "invalid_json_json_type",
        "non_list_for_list",
        "unsupported_type",
    ],
)
def test_parse_type_from_str_error_cases(input_str, output_type, expected_exception, expected_code, expected_message):
    with pytest.raises(expected_exception) as excinfo:
        parse_type_from_str(input_str, output_type)

    if expected_code:
        assert excinfo.value.code == expected_code

    assert expected_message in str(excinfo.value)


@pytest.mark.parametrize(
    "output_type,expected_result",
    [
        (str, ""),  # String
        (int, 0),  # Number
        (float, 0.0),  # Number
        (Any, None),  # Json
        (FunctionCall, FunctionCall(name="", arguments={})),  # FunctionCall
        (List[ChatMessage], []),  # Chat History
        (List[VellumValue], []),  # Array
        (Union[float, int], 0.0),  # Union
    ],
    ids=[
        "string",
        "integer",
        "float",
        "json",
        "function_call",
        "chat_history",
        "array",
        "union",
    ],
)
def test_cast_to_output_type_none_value(output_type, expected_result):
    """Test that cast_to_output_type returns appropriate default values when None is provided."""
    result = cast_to_output_type(None, output_type)
    assert result == expected_result



from vellum.workflows import BaseWorkflow
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes.displayable import SearchNode
from vellum.workflows.nodes.displayable.bases.types import (
    MetadataLogicalCondition,
    MetadataLogicalConditionGroup,
    SearchFilters,
)
from vellum.workflows.state.base import BaseState
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_search_filters_with_input_reference():
    """Test that SearchFilters with MetadataLogicalCondition using input references can be serialized"""

    # GIVEN a search node with a metadata filter that uses an input reference
    class TestInputs(BaseInputs):
        file_id: str

    class MySearchNode(SearchNode):
        query = "my query"
        document_index = "document_index"
        filters = SearchFilters(
            external_ids=None,
            metadata=MetadataLogicalConditionGroup(
                combinator="AND",
                negated=False,
                conditions=[MetadataLogicalCondition(lhs_variable="ID", operator="=", rhs_variable=TestInputs.file_id)],
            ),
        )

    # AND a workflow with the Search Node
    class Workflow(BaseWorkflow[TestInputs, BaseState]):
        graph = MySearchNode

    # WHEN the workflow is serialized
    workflow_display = get_workflow_display(workflow_class=Workflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the node should properly serialize the filter reference
    serialized_search_node = next(
        node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["id"] == str(MySearchNode.__id__)
    )

    serialized_metadata_filter = next(
        inp for inp in serialized_search_node["inputs"] if inp["key"] == "metadata_filters"
    )

    assert serialized_metadata_filter == {
        "id": "4a9f96aa-ba3b-4c4e-9ce4-370fe64f717f",
        "key": "metadata_filters",
        "value": {
            "combinator": "OR",
            "rules": [
                {
                    "data": {
                        "type": "JSON",
                        "value": {
                            "combinator": "AND",
                            "conditions": [
                                {
                                    "lhs_variable_id": "9aedaffa-c2a4-4c37-9969-184e1ff43ded",
                                    "operator": "=",
                                    "rhs_variable_id": "c2151ef1-ad98-4940-b0e9-28dabe47a951",
                                    "type": "LOGICAL_CONDITION",
                                }
                            ],
                            "negated": False,
                            "type": "LOGICAL_CONDITION_GROUP",
                        },
                    },
                    "type": "CONSTANT_VALUE",
                }
            ],
        },
    }

    # AND the LHS filter references should be present as node inputs
    serialized_lhs_input = next(
        inp for inp in serialized_search_node["inputs"] if inp["id"] == "9aedaffa-c2a4-4c37-9969-184e1ff43ded"
    )
    assert serialized_lhs_input["value"] == {
        "combinator": "OR",
        "rules": [
            {
                "data": {"type": "STRING", "value": "ID"},
                "type": "CONSTANT_VALUE",
            }
        ],
    }

    # AND the RHS filter references should be present as node inputs
    serialized_rhs_input = next(
        inp for inp in serialized_search_node["inputs"] if inp["id"] == "c2151ef1-ad98-4940-b0e9-28dabe47a951"
    )
    assert serialized_rhs_input["value"] == {
        "combinator": "OR",
        "rules": [
            {
                "data": {"input_variable_id": "e2f4fff9-1277-47cb-8988-12f8ada450ba"},
                "type": "INPUT_VARIABLE",
            }
        ],
    }



from typing import Generic, TypeVar, Union

from vellum.workflows.constants import undefined
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.descriptors.exceptions import InvalidExpressionException
from vellum.workflows.descriptors.utils import resolve_value
from vellum.workflows.errors.types import WorkflowError
from vellum.workflows.state.base import BaseState

LHS = TypeVar("LHS")
RHS = TypeVar("RHS")


class ContainsExpression(BaseDescriptor[bool], Generic[LHS, RHS]):
    def __init__(
        self,
        *,
        lhs: Union[BaseDescriptor[LHS], LHS],
        rhs: Union[BaseDescriptor[RHS], RHS],
    ) -> None:
        super().__init__(name=f"{lhs} contains {rhs}", types=(bool,))
        self._lhs = lhs
        self._rhs = rhs

    def resolve(self, state: "BaseState") -> bool:
        # Support any type that implements the in operator
        # https://app.shortcut.com/vellum/story/4658
        lhs = resolve_value(self._lhs, state)
        # assumes that lack of is also false
        if lhs is undefined:
            return False
        if not isinstance(lhs, (list, tuple, set, dict, str, WorkflowError)):
            raise InvalidExpressionException(
                f"Expected a LHS that supported `contains`, got `{lhs.__class__.__name__}`"
            )

        rhs = resolve_value(self._rhs, state)

        if isinstance(rhs, dict):
            raise InvalidExpressionException(
                "Cannot use dict as right-hand side of contains operation. "
                "Use dict keys/values or convert to strings for comparison."
            )

        return rhs in lhs



import pytest
from datetime import datetime

from click.testing import CliRunner

from vellum.client.core.api_error import ApiError
from vellum.client.types.organization_read import OrganizationRead
from vellum.client.types.workspace_read import WorkspaceRead
from vellum_cli import main as cli_main


@pytest.mark.usefixtures("info_log_level")
def test_ping__happy_path(vellum_client):
    # GIVEN a cli
    runner = CliRunner()

    # AND a valid response from both the workflow and organization API calls
    vellum_client.workspaces.workspace_identity.return_value = WorkspaceRead(
        id="1234567890",
        name="Test Workspace",
        label="Test Workspace",
        created=datetime.now(),
    )
    vellum_client.organizations.organization_identity.return_value = OrganizationRead(
        id="1234567890",
        name="Test Organization",
        new_member_join_behavior="AUTO_ACCEPT_FROM_SHARED_DOMAIN",
        allow_staff_access=True,
        limit_config={},
    )

    # WHEN calling `vellum ping`
    result = runner.invoke(cli_main, ["ping"])

    # THEN it should return Status information about the user's workspace
    assert result.exit_code == 0
    assert (
        result.output
        == """\x1b[38;20m\
Successfully authenticated with Vellum!

Organization:
    ID: 1234567890
    Name: Test Organization

Workspace:
    ID: 1234567890
    Name: Test Workspace
\x1b[0m
"""
    )


def test_ping__error_path(vellum_client):
    # GIVEN a cli
    runner = CliRunner()

    vellum_client.workspaces.workspace_identity.side_effect = ApiError(status_code=400, body="Invalid JSON")

    # WHEN the user runs the ping command
    result = runner.invoke(cli_main, ["ping"])

    # THEN the command returns an error
    assert result.exit_code == 1
    assert (
        str(result.exception)
        == "The API we tried to ping returned an invalid response. Please make sure your `VELLUM_API_URL` environment variable is set correctly."  # noqa: E501
    )


def test_ping__unauthorized_error_path(vellum_client):
    # GIVEN a cli
    runner = CliRunner()

    # GIVEN an unauthorized error with the error message from the API
    vellum_client.workspaces.workspace_identity.side_effect = ApiError(
        status_code=403, body={"detail": "Authentication credentials were not provided."}
    )

    # WHEN the user runs the ping command
    result = runner.invoke(cli_main, ["ping"])

    # THEN the command returns an error
    assert result.exit_code == 1
    assert str(result.exception) == "Please make sure your `VELLUM_API_KEY` environment variable is set correctly."



from typing import Generic, TypeVar, Union

from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.descriptors.exceptions import InvalidExpressionException
from vellum.workflows.descriptors.utils import resolve_value
from vellum.workflows.state.base import BaseState

_V = TypeVar("_V")
_S = TypeVar("_S")
_E = TypeVar("_E")


class NotBetweenExpression(BaseDescriptor[bool], Generic[_V, _S, _E]):
    def __init__(
        self,
        *,
        value: Union[BaseDescriptor[_V], _V],
        start: Union[BaseDescriptor[_S], _S],
        end: Union[BaseDescriptor[_E], _E],
    ) -> None:
        super().__init__(name=f"{value} is not between {start} and {end}", types=(bool,))
        self._value = value
        self._start = start
        self._end = end

    def resolve(self, state: "BaseState") -> bool:
        value = resolve_value(self._value, state)
        if not isinstance(value, (int, float)):
            raise InvalidExpressionException(f"Expected a numeric value, got: {value.__class__.__name__}")

        start = resolve_value(self._start, state)
        if not isinstance(start, (int, float)):
            raise InvalidExpressionException(f"Expected a numeric start value, got: {start.__class__.__name__}")

        end = resolve_value(self._end, state)
        if not isinstance(end, (int, float)):
            raise InvalidExpressionException(f"Expected a numeric end value, got: {end.__class__.__name__}")

        return value < start or value > end



import hashlib
from uuid import UUID


def generate_workflow_deployment_prefix(deployment_name: str, release_tag: str) -> str:
    """
    Generate a workflow deployment prefix from deployment name and release tag.

    Args:
        deployment_name: The name of the workflow deployment
        release_tag: The release tag to resolve

    Returns:
        The generated prefix in format vellum_workflow_deployment_{hash}
    """
    expected_hash = str(uuid4_from_hash(f"{deployment_name}|{release_tag}")).replace("-", "_")
    return f"vellum_workflow_deployment_{expected_hash}"


def uuid4_from_hash(input_str: str) -> UUID:
    # Create a SHA-256 hash of the input string
    hash_bytes = hashlib.sha256(input_str.encode()).digest()

    # Modify the hash to follow UUID4 structure
    # UUID4 has 8-4-4-4-12 hexadecimal characters
    # Version bits (4 bits) should be set to 4 (for UUID4),
    # and variant bits (2 bits) should start with 0b10

    # Convert first 16 bytes into UUID
    hash_list = list(hash_bytes[:16])

    # Set the version to 4 (UUID4)
    hash_list[6] = (hash_list[6] & 0x0F) | 0x40

    # Set the variant to 0b10xxxxxx
    hash_list[8] = (hash_list[8] & 0x3F) | 0x80

    # Create a UUID from the modified bytes
    return UUID(bytes=bytes(hash_list))



from collections import defaultdict
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import field
from datetime import datetime
import logging
from queue import Queue
from threading import Lock
from uuid import UUID, uuid4
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    SupportsIndex,
    Tuple,
    Type,
    Union,
    cast,
)
from typing_extensions import dataclass_transform

from pydantic import GetCoreSchemaHandler, ValidationInfo, field_serializer, field_validator
from pydantic_core import core_schema

from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.utils.uuid import is_valid_uuid
from vellum.workflows.constants import undefined
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.references import ExternalInputReference, OutputReference, StateValueReference
from vellum.workflows.state.delta import AppendStateDelta, SetStateDelta, StateDelta
from vellum.workflows.types.definition import CodeResourceDefinition, serialize_type_encoder_with_id
from vellum.workflows.types.generics import StateType, import_workflow_class, is_workflow_class
from vellum.workflows.types.stack import Stack
from vellum.workflows.types.utils import datetime_now, deepcopy_with_exclusions, get_class_attr_names, infer_types

if TYPE_CHECKING:
    from vellum.workflows import BaseWorkflow
    from vellum.workflows.nodes.bases import BaseNode

logger = logging.getLogger(__name__)


class _Snapshottable:
    _snapshot_callback: Callable[[Optional[StateDelta]], None]
    _path: str

    def __bind__(self, path: str, snapshot_callback: Callable[[Optional[StateDelta]], None]) -> None:
        self._snapshot_callback = snapshot_callback
        self._path = path


@dataclass_transform(kw_only_default=True)
class _BaseStateMeta(type):
    def __getattribute__(cls, name: str) -> Any:
        if not name.startswith("_"):
            instance = vars(cls).get(name, undefined)
            types = infer_types(cls, name)
            return StateValueReference(name=name, types=types, instance=instance)

        return super().__getattribute__(name)

    def __iter__(cls) -> Iterator[StateValueReference]:
        # We iterate through the inheritance hierarchy to find all the StateValueReference attached to this
        # Inputs class. __mro__ is the method resolution order, which is the order in which base classes are resolved.
        for resolved_cls in cls.__mro__:
            attr_names = get_class_attr_names(resolved_cls)
            for attr_name in attr_names:
                if attr_name == "meta":
                    continue
                attr_value = getattr(resolved_cls, attr_name)
                if not isinstance(attr_value, (StateValueReference)):
                    continue

                yield attr_value


class _SnapshottableDict(dict, _Snapshottable):
    def __setitem__(self, key: Any, value: Any) -> None:
        super().__setitem__(key, value)
        self._snapshot_callback(SetStateDelta(name=f"{self._path}.{key}", delta=value))

    def __deepcopy__(self, memo: Any) -> "_SnapshottableDict":
        y: dict = {}
        memo[id(self)] = y
        for key, value in self.items():
            y[deepcopy(key, memo)] = deepcopy(value, memo)

        y = _SnapshottableDict(y)
        y.__bind__(self._path, self._snapshot_callback)
        memo[id(self)] = y
        return y


class _SnapshottableList(list, _Snapshottable):
    def __setitem__(self, index: Union[SupportsIndex, slice], value: Any) -> None:
        super().__setitem__(index, value)
        if isinstance(index, int):
            self._snapshot_callback(SetStateDelta(name=f"{self._path}.{index}", delta=value))

    def append(self, value: Any) -> None:
        super().append(value)
        self._snapshot_callback(AppendStateDelta(name=self._path, delta=value))

    def __deepcopy__(self, memo: Any) -> "_SnapshottableList":
        y: list = []
        memo[id(self)] = y
        append = y.append
        for a in self:
            append(deepcopy(a, memo))

        y = _SnapshottableList(y)
        y.__bind__(self._path, self._snapshot_callback)
        memo[id(self)] = y
        return y


def _make_snapshottable(path: str, value: Any, snapshot_callback: Callable[[Optional[StateDelta]], None]) -> Any:
    """
    Edits any value to make it snapshottable on edit. Made as a separate function from `BaseState` to
    avoid namespace conflicts with subclasses.
    """
    if isinstance(value, _Snapshottable):
        return value

    if isinstance(value, dict):
        snapshottable_dict = _SnapshottableDict(value)
        snapshottable_dict.__bind__(path, snapshot_callback)
        return snapshottable_dict

    if isinstance(value, list):
        snapshottable_list = _SnapshottableList(value)
        snapshottable_list.__bind__(path, snapshot_callback)
        return snapshottable_list

    return value


NodeExecutionsFulfilled = Dict[Type["BaseNode"], Stack[UUID]]
NodeExecutionsInitiated = Dict[Type["BaseNode"], Set[UUID]]
NodeExecutionsQueued = Dict[Type["BaseNode"], List[UUID]]
NodeExecutionLookup = Dict[UUID, Type["BaseNode"]]
DependenciesInvoked = Dict[UUID, Set[UUID]]


class NodeExecutionCache:
    _node_executions_fulfilled: NodeExecutionsFulfilled
    _node_executions_initiated: NodeExecutionsInitiated
    _node_executions_queued: NodeExecutionsQueued
    _dependencies_invoked: DependenciesInvoked

    # Derived fields - no need to serialize
    __node_execution_lookup__: NodeExecutionLookup  # execution_id -> node_class

    def __init__(self) -> None:
        self._dependencies_invoked = defaultdict(set)
        self._node_executions_fulfilled = defaultdict(Stack[UUID])
        self._node_executions_initiated = defaultdict(set)
        self._node_executions_queued = defaultdict(list)
        self.__node_execution_lookup__ = {}

    @classmethod
    def deserialize(cls, raw_data: dict, nodes: Dict[Union[str, UUID], Type["BaseNode"]]):
        cache = cls()

        def get_node_class(node_id: Any) -> Optional[Type["BaseNode"]]:
            if not isinstance(node_id, str):
                return None

            if is_valid_uuid(node_id):
                return nodes.get(UUID(node_id))

            return nodes.get(node_id)

        dependencies_invoked = raw_data.get("dependencies_invoked")
        if isinstance(dependencies_invoked, dict):
            for execution_id, dependencies in dependencies_invoked.items():
                dependency_execution_ids = {UUID(dep) for dep in dependencies if is_valid_uuid(dep)}
                cache._dependencies_invoked[UUID(execution_id)] = dependency_execution_ids

        node_executions_fulfilled = raw_data.get("node_executions_fulfilled")
        if isinstance(node_executions_fulfilled, dict):
            for node, execution_ids in node_executions_fulfilled.items():
                node_class = get_node_class(node)
                if not node_class:
                    continue

                cache._node_executions_fulfilled[node_class].extend(
                    UUID(execution_id) for execution_id in execution_ids
                )

        node_executions_initiated = raw_data.get("node_executions_initiated")
        if isinstance(node_executions_initiated, dict):
            for node, execution_ids in node_executions_initiated.items():
                node_class = get_node_class(node)
                if not node_class:
                    continue

                cache._node_executions_initiated[node_class].update(
                    {UUID(execution_id) for execution_id in execution_ids}
                )

            for node_class, execution_ids in cache._node_executions_initiated.items():
                for execution_id in execution_ids:
                    cache.__node_execution_lookup__[execution_id] = node_class

        node_executions_queued = raw_data.get("node_executions_queued")
        if isinstance(node_executions_queued, dict):
            for node, execution_ids in node_executions_queued.items():
                node_class = get_node_class(node)
                if not node_class:
                    continue

                cache._node_executions_queued[node_class].extend(UUID(execution_id) for execution_id in execution_ids)

        return cache

    def _invoke_dependency(
        self,
        execution_id: UUID,
        node: Type["BaseNode"],
        invoked_by: UUID,
        dependencies: Set["Type[BaseNode]"],
    ) -> None:
        self._dependencies_invoked[execution_id].add(invoked_by)
        invoked_node_classes = {
            self.__node_execution_lookup__[dep]
            for dep in self._dependencies_invoked[execution_id]
            if dep in self.__node_execution_lookup__
        }
        if len(invoked_node_classes) != len(dependencies):
            return

        if any(dep not in invoked_node_classes for dep in dependencies):
            return

        self._node_executions_queued[node].remove(execution_id)

    def is_node_execution_initiated(self, node: Type["BaseNode"], execution_id: UUID) -> bool:
        return execution_id in self._node_executions_initiated[node]

    def initiate_node_execution(self, node: Type["BaseNode"], execution_id: UUID) -> None:
        self._node_executions_initiated[node].add(execution_id)
        self.__node_execution_lookup__[execution_id] = node

    def fulfill_node_execution(self, node: Type["BaseNode"], execution_id: UUID) -> None:
        self._node_executions_fulfilled[node].push(execution_id)

    def get_execution_count(self, node: Type["BaseNode"]) -> int:
        return self._node_executions_fulfilled[node].size()

    def dump(self) -> Dict[str, Any]:
        return {
            "dependencies_invoked": {
                str(execution_id): [str(dep) for dep in dependencies]
                for execution_id, dependencies in self._dependencies_invoked.items()
            },
            "node_executions_initiated": {
                str(node.__id__): list(execution_ids) for node, execution_ids in self._node_executions_initiated.items()
            },
            "node_executions_fulfilled": {
                str(node.__id__): execution_ids.dump()
                for node, execution_ids in self._node_executions_fulfilled.items()
            },
            "node_executions_queued": {
                str(node.__id__): execution_ids for node, execution_ids in self._node_executions_queued.items()
            },
        }

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.is_instance_schema(cls)


def uuid4_default_factory() -> UUID:
    """
    Allows us to mock the uuid4 for testing.
    """
    return uuid4()


def default_datetime_factory() -> datetime:
    """
    Makes it possible to mock the datetime factory for testing.
    """

    return datetime_now()


class StateMeta(UniversalBaseModel):
    workflow_definition: Type["BaseWorkflow"] = field(default_factory=import_workflow_class)
    id: UUID = field(default_factory=uuid4_default_factory)
    span_id: UUID = field(default_factory=uuid4_default_factory)
    updated_ts: datetime = field(default_factory=default_datetime_factory)
    workflow_inputs: BaseInputs = field(default_factory=BaseInputs)
    external_inputs: Dict[ExternalInputReference, Any] = field(default_factory=dict)
    node_outputs: Dict[OutputReference, Any] = field(default_factory=dict)
    node_execution_cache: NodeExecutionCache = field(default_factory=NodeExecutionCache)
    parent: Optional["BaseState"] = None
    __snapshot_callback__: Optional[Callable[[Optional[StateDelta]], None]] = field(init=False, default=None)

    def model_post_init(self, context: Any) -> None:
        self.__snapshot_callback__ = None

    def add_snapshot_callback(self, callback: Callable[[Optional[StateDelta]], None]) -> None:
        self.node_outputs = _make_snapshottable("meta.node_outputs", self.node_outputs, callback)
        self.external_inputs = _make_snapshottable("meta.external_inputs", self.external_inputs, callback)
        self.__snapshot_callback__ = callback

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("__") or name == "updated_ts":
            super().__setattr__(name, value)
            return

        super().__setattr__(name, value)
        if callable(self.__snapshot_callback__):
            self.__snapshot_callback__(SetStateDelta(name=f"meta.{name}", delta=value))

    @field_serializer("workflow_definition")
    def serialize_workflow_definition(self, workflow_definition: Type["BaseWorkflow"], _info: Any) -> Dict[str, Any]:
        return serialize_type_encoder_with_id(workflow_definition)

    @field_validator("workflow_definition", mode="before")
    @classmethod
    def deserialize_workflow_definition(cls, workflow_definition: Any, info: ValidationInfo):
        if isinstance(workflow_definition, dict):
            deserialized_workflow_definition = CodeResourceDefinition.model_validate(workflow_definition).decode()
            if not is_workflow_class(deserialized_workflow_definition):
                return import_workflow_class()

            return deserialized_workflow_definition

        if is_workflow_class(workflow_definition):
            return workflow_definition

        return import_workflow_class()

    @field_serializer("node_outputs")
    def serialize_node_outputs(self, node_outputs: Dict[OutputReference, Any], _info: Any) -> Dict[str, Any]:
        return {str(descriptor.id): value for descriptor, value in node_outputs.items()}

    @field_validator("node_outputs", mode="before")
    @classmethod
    def deserialize_node_outputs(cls, node_outputs: Any, info: ValidationInfo):
        if isinstance(node_outputs, dict):
            workflow_definition = cls._get_workflow(info)
            if not workflow_definition:
                return node_outputs

            raw_workflow_nodes = workflow_definition.get_nodes()
            workflow_node_outputs: Dict[Union[str, UUID], OutputReference] = {}
            for node in raw_workflow_nodes:
                for output in node.Outputs:
                    workflow_node_outputs[str(output)] = output
                    output_id = node.__output_ids__.get(output.name)
                    if output_id:
                        workflow_node_outputs[output_id] = output

            node_output_keys = list(node_outputs.keys())
            deserialized_node_outputs = {}
            for node_output_key in node_output_keys:
                if is_valid_uuid(node_output_key):
                    output_reference = workflow_node_outputs.get(UUID(node_output_key))
                else:
                    output_reference = workflow_node_outputs.get(node_output_key)

                if not output_reference:
                    continue

                deserialized_node_outputs[output_reference] = node_outputs[node_output_key]

            return deserialized_node_outputs

        return node_outputs

    @field_validator("node_execution_cache", mode="before")
    @classmethod
    def deserialize_node_execution_cache(cls, node_execution_cache: Any, info: ValidationInfo):
        if isinstance(node_execution_cache, dict):
            workflow_definition = cls._get_workflow(info)
            if not workflow_definition:
                return node_execution_cache

            nodes_cache: Dict[Union[str, UUID], Type["BaseNode"]] = {}
            raw_workflow_nodes = workflow_definition.get_nodes()
            for node in raw_workflow_nodes:
                nodes_cache[str(node)] = node
                nodes_cache[node.__id__] = node

            return NodeExecutionCache.deserialize(node_execution_cache, nodes_cache)

        return node_execution_cache

    @field_validator("workflow_inputs", mode="before")
    @classmethod
    def deserialize_workflow_inputs(cls, workflow_inputs: Any, info: ValidationInfo):
        workflow_definition = cls._get_workflow(info)

        if workflow_definition:
            if workflow_inputs is None:
                return workflow_definition.get_inputs_class()()
            if isinstance(workflow_inputs, dict):
                return workflow_definition.get_inputs_class()(**workflow_inputs)

        return workflow_inputs

    @field_serializer("external_inputs")
    def serialize_external_inputs(
        self, external_inputs: Dict[ExternalInputReference, Any], _info: Any
    ) -> Dict[str, Any]:
        return {str(descriptor): value for descriptor, value in external_inputs.items()}

    @field_validator("parent", mode="before")
    @classmethod
    def deserialize_parent(cls, parent: Any, info: ValidationInfo):
        if isinstance(parent, dict):
            workflow_definition = cls._get_workflow(info)
            if not workflow_definition:
                return parent

            parent_meta = parent.get("meta")
            if not isinstance(parent_meta, dict):
                return parent

            parent_workflow_definition = cls.deserialize_workflow_definition(
                parent_meta.get("workflow_definition"), info
            )
            if not is_workflow_class(parent_workflow_definition):
                return parent

            return parent_workflow_definition.deserialize_state(parent)

        return parent

    def __deepcopy__(self, memo: Optional[Dict[int, Any]] = None) -> "StateMeta":
        if not memo:
            memo = {}

        new_node_outputs = {
            descriptor: value if isinstance(value, Queue) else deepcopy(value, memo)
            for descriptor, value in self.node_outputs.items()
        }

        new_external_inputs = {
            descriptor: value if isinstance(value, Queue) else deepcopy(value, memo)
            for descriptor, value in self.external_inputs.items()
        }

        memo[id(self.node_outputs)] = new_node_outputs
        memo[id(self.external_inputs)] = new_external_inputs
        memo[id(self.__snapshot_callback__)] = None

        return super().__deepcopy__(memo)

    @classmethod
    def _get_workflow(cls, info: ValidationInfo) -> Optional[Type["BaseWorkflow"]]:
        if not info.context:
            return None

        if not isinstance(info.context, dict):
            return None

        workflow_definition = info.context.get("workflow_definition")
        if not is_workflow_class(workflow_definition):
            return None

        return workflow_definition


class BaseState(metaclass=_BaseStateMeta):
    meta: StateMeta = field(init=False)

    __lock__: Lock = field(init=False)
    __is_quiet__: bool = field(init=False)
    __is_atomic__: bool = field(init=False)
    __snapshot_callback__: Callable[["BaseState", List[StateDelta]], None] = field(init=False)
    __deltas__: List[StateDelta] = field(init=False)

    def __init__(self, meta: Optional[StateMeta] = None, **kwargs: Any) -> None:
        self.__is_quiet__ = True
        self.__is_atomic__ = False
        self.__snapshot_callback__ = lambda state, deltas: None
        self.__deltas__ = []
        self.__lock__ = Lock()

        self.meta = meta or StateMeta()
        self.meta.add_snapshot_callback(self.__snapshot__)

        # Make all class attribute values snapshottable
        for name, value in self.__class__.__dict__.items():
            if not name.startswith("_") and name != "meta":
                # Bypass __is_quiet__ instead of `setattr`
                snapshottable_value = _make_snapshottable(name, value, self.__snapshot__)
                super().__setattr__(name, snapshottable_value)

        for name, value in kwargs.items():
            setattr(self, name, value)

        self.__is_quiet__ = False

    def __deepcopy__(self, memo: Any) -> "BaseState":
        new_state = deepcopy_with_exclusions(
            self,
            exclusions={
                "__lock__": Lock(),
            },
            memo=memo,
        )
        new_state.meta.add_snapshot_callback(new_state.__snapshot__)
        return new_state

    def __repr__(self) -> str:
        values = "\n".join(
            [f"    {key}={value}" for key, value in vars(self).items() if not key.startswith("_") and key != "meta"]
        )
        node_outputs = "\n".join([f"            {key}={value}" for key, value in self.meta.node_outputs.items()])
        return f"""\
{self.__class__.__name__}:
{values}
    meta:
        id={self.meta.id}
        updated_ts={self.meta.updated_ts}
        node_outputs:{' Empty' if not node_outputs else ''}
{node_outputs}
"""

    def __iter__(self) -> Iterator[Tuple[Any, Any]]:
        """
        Returns an iterator treating all state keys as (key, value) items, allowing consumers to call `dict()`
        on an instance of this class.
        """

        # If the user sets a default value on state (e.g. something = "foo"), it's not on `instance_attributes` below.
        # So we need to include class_attributes here just in case
        class_attributes = {key: value for key, value in self.__class__.__dict__.items() if not key.startswith("_")}
        instance_attributes = {key: value for key, value in self.__dict__.items() if not key.startswith("__")}

        all_attributes = {**class_attributes, **instance_attributes}
        items = [(key, value) for key, value in all_attributes.items() if key not in ["_lock"]]
        return iter(items)

    def __getitem__(self, key: str) -> Any:
        return self.__dict__[key]

    def __setattr__(self, name: str, delta: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, delta)
            return

        snapshottable_value = _make_snapshottable(name, delta, self.__snapshot__)
        super().__setattr__(name, snapshottable_value)
        self.meta.updated_ts = datetime_now()
        self.__snapshot__(SetStateDelta(name=name, delta=delta))

    def __add__(self, other: StateType) -> StateType:
        """
        Handles merging two states together, preferring the latest state by updated_ts for any given node output.
        """

        if not isinstance(other, type(self)):
            raise TypeError(f"Cannot add {type(other).__name__} to {type(self).__name__}]")

        latest_state = self if self.meta.updated_ts >= other.meta.updated_ts else other
        oldest_state = other if latest_state == self else self

        for descriptor, value in oldest_state.meta.node_outputs.items():
            if descriptor not in latest_state.meta.node_outputs:
                latest_state.meta.node_outputs[descriptor] = value

        for key, value in oldest_state:
            if not isinstance(key, str):
                continue

            if key.startswith("_"):
                continue

            if getattr(latest_state, key, undefined) == undefined:
                setattr(latest_state, key, value)

        return cast(StateType, latest_state)

    def __snapshot__(self, delta: Optional[StateDelta] = None) -> None:
        """
        Snapshots the current state to the workflow emitter. The invoked callback is overridden by the
        workflow runner.
        """
        if self.__is_quiet__:
            return

        if delta:
            self.__deltas__.append(delta)

        if self.__is_atomic__:
            return

        try:
            self.__snapshot_callback__(deepcopy(self), self.__deltas__)
        except Exception:
            logger.exception("Failed to snapshot Workflow state.")

        self.__deltas__.clear()

    @contextmanager
    def __quiet__(self):
        prev = self.__is_quiet__
        self.__is_quiet__ = True
        try:
            yield
        finally:
            self.__is_quiet__ = prev

    @contextmanager
    def __atomic__(self):
        prev = self.__is_atomic__
        self.__is_atomic__ = True
        try:
            yield
        finally:
            self.__is_atomic__ = prev
            self.__snapshot__()

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.is_instance_schema(cls)
